@@ -1,4 +1,4 @@
-# Copyright 2019-2020 Canonical Ltd.
+# Copyright 2020 Canonical Ltd.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,18 +16,23 @@ import collections
 import collections.abc
 import inspect
 import keyword
+import logging
 import marshal
 import os
+import pathlib
 import pdb
-import pickle
 import re
-import sqlite3
 import sys
 import types
 import weakref
-from datetime import timedelta
 
 from ops import charm
+from ops.storage import (
+    NoSnapshotError,
+    SQLiteStorage,
+)
+
+logger = logging.getLogger(__name__)
 
 
 class Handle:
@@ -115,7 +120,11 @@ class EventBase:
         self.handle = handle
         self.deferred = False
 
+    def __repr__(self):
+        return "<%s via %s>" % (self.__class__.__name__, self.handle)
+
     def defer(self):
+        logger.debug("Deferring %s.", self)
         self.deferred = True
 
     def snapshot(self):
@@ -369,15 +378,6 @@ class FrameworkEvents(ObjectEvents):
     commit = EventSource(CommitEvent)
 
 
-class NoSnapshotError(Exception):
-
-    def __init__(self, handle_path):
-        self.handle_path = handle_path
-
-    def __str__(self):
-        return 'no snapshot data found for {} object'.format(self.handle_path)
-
-
 class NoTypeError(Exception):
 
     def __init__(self, handle_path):
@@ -387,97 +387,6 @@ class NoTypeError(Exception):
         return "cannot restore {} since no class was registered for it".format(self.handle_path)
 
 
-class SQLiteStorage:
-
-    DB_LOCK_TIMEOUT = timedelta(hours=1)
-
-    def __init__(self, filename):
-        # The isolation_level argument is set to None such that the implicit
-        # transaction management behavior of the sqlite3 module is disabled.
-        self._db = sqlite3.connect(str(filename),
-                                   isolation_level=None,
-                                   timeout=self.DB_LOCK_TIMEOUT.total_seconds())
-        self._setup()
-
-    def _setup(self):
-        # Make sure that the database is locked until the connection is closed,
-        # not until the transaction ends.
-        self._db.execute("PRAGMA locking_mode=EXCLUSIVE")
-        c = self._db.execute("BEGIN")
-        c.execute("SELECT count(name) FROM sqlite_master WHERE type='table' AND name='snapshot'")
-        if c.fetchone()[0] == 0:
-            # Keep in mind what might happen if the process dies somewhere below.
-            # The system must not be rendered permanently broken by that.
-            self._db.execute("CREATE TABLE snapshot (handle TEXT PRIMARY KEY, data BLOB)")
-            self._db.execute('''
-                CREATE TABLE notice (
-                  sequence INTEGER PRIMARY KEY AUTOINCREMENT,
-                  event_path TEXT,
-                  observer_path TEXT,
-                  method_name TEXT)
-                ''')
-            self._db.commit()
-
-    def close(self):
-        self._db.close()
-
-    def commit(self):
-        self._db.commit()
-
-    # There's commit but no rollback. For abort to be supported, we'll need logic that
-    # can rollback decisions made by third-party code in terms of the internal state
-    # of objects that have been snapshotted, and hooks to let them know about it and
-    # take the needed actions to undo their logic until the last snapshot.
-    # This is doable but will increase significantly the chances for mistakes.
-
-    def save_snapshot(self, handle_path, snapshot_data):
-        self._db.execute("REPLACE INTO snapshot VALUES (?, ?)", (handle_path, snapshot_data))
-
-    def load_snapshot(self, handle_path):
-        c = self._db.cursor()
-        c.execute("SELECT data FROM snapshot WHERE handle=?", (handle_path,))
-        row = c.fetchone()
-        if row:
-            return row[0]
-        return None
-
-    def drop_snapshot(self, handle_path):
-        self._db.execute("DELETE FROM snapshot WHERE handle=?", (handle_path,))
-
-    def save_notice(self, event_path, observer_path, method_name):
-        self._db.execute('INSERT INTO notice VALUES (NULL, ?, ?, ?)',
-                         (event_path, observer_path, method_name))
-
-    def drop_notice(self, event_path, observer_path, method_name):
-        self._db.execute('''
-            DELETE FROM notice
-             WHERE event_path=?
-               AND observer_path=?
-               AND method_name=?
-            ''', (event_path, observer_path, method_name))
-
-    def notices(self, event_path):
-        if event_path:
-            c = self._db.execute('''
-                SELECT event_path, observer_path, method_name
-                  FROM notice
-                 WHERE event_path=?
-                 ORDER BY sequence
-                ''', (event_path,))
-        else:
-            c = self._db.execute('''
-                SELECT event_path, observer_path, method_name
-                  FROM notice
-                 ORDER BY sequence
-                ''')
-        while True:
-            rows = c.fetchmany()
-            if not rows:
-                break
-            for row in rows:
-                yield tuple(row)
-
-
 # the message to show to the user when a pdb breakpoint goes active
 _BREAKPOINT_WELCOME_MESSAGE = """
 Starting pdb to debug charm operator.
@@ -488,6 +397,9 @@ More details at https://discourse.jujucharms.com/t/debugging-charm-hooks
 """
 
 
+_event_regex = r'^(|.*/)on/[a-zA-Z_]+\[\d+\]$'
+
+
 class Framework(Object):
 
     on = FrameworkEvents()
@@ -497,11 +409,10 @@ class Framework(Object):
     meta = None
     charm_dir = None
 
-    def __init__(self, data_path, charm_dir, meta, model):
+    def __init__(self, storage, charm_dir, meta, model):
 
         super().__init__(self, None)
 
-        self._data_path = data_path
         self.charm_dir = charm_dir
         self.meta = meta
         self.model = model
@@ -511,7 +422,11 @@ class Framework(Object):
         self._type_registry = {}  # {(parent_path, kind): cls}
         self._type_known = set()  # {cls}
 
-        self._storage = SQLiteStorage(data_path)
+        if isinstance(storage, (str, pathlib.Path)):
+            logger.warning(
+                "deprecated: Framework now takes a Storage not a path")
+            storage = SQLiteStorage(storage)
+        self._storage = storage
 
         # We can't use the higher-level StoredState because it relies on events.
         self.register_type(StoredStateData, None, StoredStateData.handle_kind)
@@ -597,9 +512,7 @@ class Framework(Object):
             msg = "unable to save the data for {}, it must contain only simple types: {!r}"
             raise ValueError(msg.format(value.__class__.__name__, data))
 
-        # Use pickle for serialization, so the value remains portable.
-        raw_data = pickle.dumps(data)
-        self._storage.save_snapshot(value.handle.path, raw_data)
+        self._storage.save_snapshot(value.handle.path, data)
 
     def load_snapshot(self, handle):
         parent_path = None
@@ -608,10 +521,7 @@ class Framework(Object):
         cls = self._type_registry.get((parent_path, handle.kind))
         if not cls:
             raise NoTypeError(handle.path)
-        raw_data = self._storage.load_snapshot(handle.path)
-        if not raw_data:
-            raise NoSnapshotError(handle.path)
-        data = pickle.loads(raw_data)
+        data = self._storage.load_snapshot(handle.path)
         obj = cls.__new__(cls)
         obj.framework = self
         obj.handle = handle
@@ -696,9 +606,7 @@ class Framework(Object):
     def _emit(self, event):
         """See BoundEvent.emit for the public way to call this."""
 
-        # Save the event for all known observers before the first notification
-        # takes place, so that either everyone interested sees it, or nobody does.
-        self.save_snapshot(event)
+        saved = False
         event_path = event.handle.path
         event_kind = event.handle.kind
         parent_path = event.handle.parent.path
@@ -709,9 +617,15 @@ class Framework(Object):
                 continue
             if _event_kind and _event_kind != event_kind:
                 continue
+            if not saved:
+                # Save the event for all known observers before the first notification
+                # takes place, so that either everyone interested sees it, or nobody does.
+                self.save_snapshot(event)
+                saved = True
             # Again, only commit this after all notices are saved.
             self._storage.save_notice(event_path, observer_path, method_name)
-        self._reemit(event_path)
+        if saved:
+            self._reemit(event_path)
 
     def reemit(self):
         """Reemit previously deferred events to the observers that deferred them.
@@ -730,7 +644,7 @@ class Framework(Object):
             event_handle = Handle.from_path(event_path)
 
             if last_event_path != event_path:
-                if not deferred:
+                if not deferred and last_event_path is not None:
                     self._storage.drop_snapshot(last_event_path)
                 last_event_path = event_path
                 deferred = False
@@ -744,6 +658,8 @@ class Framework(Object):
             event.deferred = False
             observer = self._observer.get(observer_path)
             if observer:
+                if single_event_path is None:
+                    logger.debug("Re-emitting %s.", event)
                 custom_handler = getattr(observer, method_name, None)
                 if custom_handler:
                     event_is_from_juju = isinstance(event, charm.HookEvent)
@@ -764,7 +680,7 @@ class Framework(Object):
             # scratch in the next path.
             self.framework._forget(event)
 
-        if not deferred:
+        if not deferred and last_event_path is not None:
             self._storage.drop_snapshot(last_event_path)
 
     def _show_debug_code_message(self):
@@ -795,6 +711,9 @@ class Framework(Object):
                 raise ValueError('breakpoint names must look like "foo" or "foo-bar"')
 
         indicated_breakpoints = self._juju_debug_at
+        if not indicated_breakpoints:
+            return
+
         if 'all' in indicated_breakpoints or name in indicated_breakpoints:
             self._show_debug_code_message()
 
@@ -802,6 +721,28 @@ class Framework(Object):
             # it to use our caller's frame
             code_frame = inspect.currentframe().f_back
             pdb.Pdb().set_trace(code_frame)
+        else:
+            logger.warning(
+                "Breakpoint %r skipped (not found in the requested breakpoints: %s)",
+                name, indicated_breakpoints)
+
+    def remove_unreferenced_events(self):
+        """Remove events from storage that are not referenced.
+
+        In older versions of the framework, events that had no observers would get recorded but
+        never deleted. This makes a best effort to find these events and remove them from the
+        database.
+        """
+        event_regex = re.compile(_event_regex)
+        to_remove = []
+        for handle_path in self._storage.list_snapshots():
+            if event_regex.match(handle_path):
+                notices = self._storage.notices(handle_path)
+                if next(notices, None) is None:
+                    # There are no notices for this handle_path, it is valid to remove it
+                    to_remove.append(handle_path)
+        for handle_path in to_remove:
+            self._storage.drop_snapshot(handle_path)
 
 
 class StoredStateData(Object):
