@@ -28,43 +28,79 @@ from abc import ABC, abstractmethod
 from collections.abc import Mapping, MutableMapping
 from pathlib import Path
 from subprocess import run, PIPE, CalledProcessError
+import yaml
 
 import ops
 from ops.jujuversion import JujuVersion
 
 
+if yaml.__with_libyaml__:
+    _DefaultDumper = yaml.CSafeDumper
+else:
+    _DefaultDumper = yaml.SafeDumper
+
+
 class Model:
     """Represents the Juju Model as seen from this unit.
 
     This should not be instantiated directly by Charmers, but can be accessed as `self.model`
     from any class that derives from Object.
-
-    Attributes:
-        unit: A :class:`Unit` that represents the unit that is running this code (eg yourself)
-        app: A :class:`Application` that represents the application this unit is a part of.
-        relations: Mapping of endpoint to list of :class:`Relation` answering the question
-            "what am I currently related to". See also :meth:`.get_relation`
-        config: A dict of the config for the current application.
-        resources: Access to resources for this charm. Use ``model.resources.fetch(resource_name)``
-            to get the path on disk where the resource can be found.
-        storages: Mapping of storage_name to :class:`Storage` for the storage points defined in
-            metadata.yaml
-        pod: Used to get access to ``model.pod.set_spec`` to set the container specification
-            for Kubernetes charms.
     """
 
     def __init__(self, meta: 'ops.charm.CharmMeta', backend: '_ModelBackend'):
         self._cache = _ModelCache(backend)
         self._backend = backend
-        self.unit = self.get_unit(self._backend.unit_name)
-        self.app = self.unit.app
-        self.relations = RelationMapping(meta.relations, self.unit, self._backend, self._cache)
-        self.config = ConfigData(self._backend)
-        self.resources = Resources(list(meta.resources), self._backend)
-        self.pod = Pod(self._backend)
-        self.storages = StorageMapping(list(meta.storages), self._backend)
+        self._unit = self.get_unit(self._backend.unit_name)
+        self._relations = RelationMapping(meta.relations, self.unit, self._backend, self._cache)
+        self._config = ConfigData(self._backend)
+        self._resources = Resources(list(meta.resources), self._backend)
+        self._pod = Pod(self._backend)
+        self._storages = StorageMapping(list(meta.storages), self._backend)
         self._bindings = BindingMapping(self._backend)
 
+    @property
+    def unit(self) -> 'Unit':
+        """A :class:`Unit` that represents the unit that is running this code (eg yourself)"""
+        return self._unit
+
+    @property
+    def app(self):
+        """A :class:`Application` that represents the application this unit is a part of."""
+        return self._unit.app
+
+    @property
+    def relations(self) -> 'RelationMapping':
+        """Mapping of endpoint to list of :class:`Relation`
+
+        Answers the question "what am I currently related to".
+        See also :meth:`.get_relation`.
+        """
+        return self._relations
+
+    @property
+    def config(self) -> 'ConfigData':
+        """Return a mapping of config for the current application."""
+        return self._config
+
+    @property
+    def resources(self) -> 'Resources':
+        """Access to resources for this charm.
+
+        Use ``model.resources.fetch(resource_name)`` to get the path on disk
+        where the resource can be found.
+        """
+        return self._resources
+
+    @property
+    def storages(self) -> 'StorageMapping':
+        """Mapping of storage_name to :class:`Storage` as defined in metadata.yaml"""
+        return self._storages
+
+    @property
+    def pod(self) -> 'Pod':
+        """Use ``model.pod.set_spec`` to set the container specification for Kubernetes charms."""
+        return self._pod
+
     @property
     def name(self) -> str:
         """Return the name of the Model that this unit is running in.
@@ -341,6 +377,9 @@ class LazyMapping(Mapping, ABC):
     def __getitem__(self, key):
         return self._data[key]
 
+    def __repr__(self):
+        return repr(self._data)
+
 
 class RelationMapping(Mapping):
     """Map of relation names to lists of :class:`Relation` instances."""
@@ -480,7 +519,7 @@ class Network:
             information about how your application should be configured (eg, what
             IP addresses should you bind to.)
             Note that multiple addresses for a single interface are represented as multiple
-            interfaces. (eg, ``[NetworKInfo('ens1', '10.1.1.1/32'),
+            interfaces. (eg, ``[NetworkInfo('ens1', '10.1.1.1/32'),
             NetworkInfo('ens1', '10.1.2.1/32'])``)
         ingress_addresses: A list of :class:`ipaddress.ip_address` objects representing the IP
             addresses that other units should use to get in touch with you.
@@ -641,6 +680,9 @@ class RelationData(Mapping):
     def __getitem__(self, key):
         return self._data[key]
 
+    def __repr__(self):
+        return repr(self._data)
+
 
 # We mix in MutableMapping here to get some convenience implementations, but whether it's actually
 # mutable or not is controlled by the flag.
@@ -687,7 +729,7 @@ class RelationDataContent(LazyMapping, MutableMapping):
             if value == '':
                 # Match the behavior of Juju, which is that setting the value to an
                 # empty string will remove the key entirely from the relation data.
-                del self._data[key]
+                self._data.pop(key, None)
             else:
                 self._data[key] = value
 
@@ -844,11 +886,13 @@ class Pod:
         """Set the specification for pods that Juju should start in kubernetes.
 
         See `juju help-tool pod-spec-set` for details of what should be passed.
+
         Args:
             spec: The mapping defining the pod specification
             k8s_resources: Additional kubernetes specific specification.
 
         Returns:
+            None
         """
         if not self._backend.is_leader():
             raise ModelError('cannot set a pod spec as this unit is not a leader')
@@ -969,11 +1013,12 @@ class _ModelBackend:
         self._leader_check_time = None
 
     def _run(self, *args, return_output=False, use_json=False):
-        kwargs = dict(stdout=PIPE, stderr=PIPE)
+        kwargs = dict(stdout=PIPE, stderr=PIPE, check=True)
+        args = (shutil.which(args[0]),) + args[1:]
         if use_json:
             args += ('--format=json',)
         try:
-            result = run(args, check=True, **kwargs)
+            result = run(args, **kwargs)
         except CalledProcessError as e:
             raise ModelError(e.stderr)
         if return_output:
@@ -1069,12 +1114,14 @@ class _ModelBackend:
     def pod_spec_set(self, spec, k8s_resources):
         tmpdir = Path(tempfile.mkdtemp('-pod-spec-set'))
         try:
-            spec_path = tmpdir / 'spec.json'
-            spec_path.write_text(json.dumps(spec))
+            spec_path = tmpdir / 'spec.yaml'
+            with spec_path.open("wt", encoding="utf8") as f:
+                yaml.dump(spec, stream=f, Dumper=_DefaultDumper)
             args = ['--file', str(spec_path)]
             if k8s_resources:
-                k8s_res_path = tmpdir / 'k8s-resources.json'
-                k8s_res_path.write_text(json.dumps(k8s_resources))
+                k8s_res_path = tmpdir / 'k8s-resources.yaml'
+                with k8s_res_path.open("wt", encoding="utf8") as f:
+                    yaml.dump(k8s_resources, stream=f, Dumper=_DefaultDumper)
                 args.extend(['--k8s-resources', str(k8s_res_path)])
             self._run('pod-spec-set', *args)
         finally:
