@@ -17,29 +17,33 @@ import gc
 import inspect
 import io
 import os
+import re
 import shutil
 import sys
 import tempfile
 from unittest.mock import patch
 from pathlib import Path
 
+import logassert
+
 from ops import charm
 from ops.framework import (
     _BREAKPOINT_WELCOME_MESSAGE,
     BoundStoredState,
     CommitEvent,
     EventBase,
+    _event_regex,
     ObjectEvents,
     EventSource,
     Framework,
     Handle,
-    NoSnapshotError,
     Object,
     PreCommitEvent,
     StoredList,
     StoredState,
     StoredStateData,
 )
+from ops.storage import NoSnapshotError, SQLiteStorage
 from test.test_helpers import fake_script, BaseTestCase
 
 
@@ -49,9 +53,17 @@ class TestFramework(BaseTestCase):
         self.tmpdir = Path(tempfile.mkdtemp())
         self.addCleanup(shutil.rmtree, str(self.tmpdir))
 
-        patcher = patch('ops.framework.SQLiteStorage.DB_LOCK_TIMEOUT', datetime.timedelta(0))
+        patcher = patch('ops.storage.SQLiteStorage.DB_LOCK_TIMEOUT', datetime.timedelta(0))
         patcher.start()
         self.addCleanup(patcher.stop)
+        logassert.setup(self, 'ops')
+
+    def test_deprecated_init(self):
+        # For 0.7, this still works, but it is deprecated.
+        framework = Framework(':memory:', None, None, None)
+        self.assertLoggedWarning(
+            "deprecated: Framework now takes a Storage not a path")
+        self.assertIsInstance(framework._storage, SQLiteStorage)
 
     def test_handle_path(self):
         cases = [
@@ -149,12 +161,15 @@ class TestFramework(BaseTestCase):
             def __init__(self, parent, key):
                 super().__init__(parent, key)
                 self.seen = []
+                self.reprs = []
 
             def on_any(self, event):
                 self.seen.append("on_any:" + event.handle.kind)
+                self.reprs.append(repr(event))
 
             def on_foo(self, event):
                 self.seen.append("on_foo:" + event.handle.kind)
+                self.reprs.append(repr(event))
 
         pub = MyNotifier(framework, "1")
         obs = MyObserver(framework, "1")
@@ -169,6 +184,10 @@ class TestFramework(BaseTestCase):
         pub.bar.emit()
 
         self.assertEqual(obs.seen, ["on_any:foo", "on_any:bar"])
+        self.assertEqual(obs.reprs, [
+            "<MyEvent via MyNotifier[1]/foo[1]>",
+            "<MyEvent via MyNotifier[1]/bar[2]>",
+        ])
 
     def test_bad_sig_observer(self):
 
@@ -353,7 +372,10 @@ class TestFramework(BaseTestCase):
 
         framework.observe(pub.foo, obs._on_foo)
 
+        self.assertNotLogged("Deferring")
         pub.foo.emit(1)
+        self.assertLogged("Deferring <MyEvent via MyNotifier[1]/foo[1]>.")
+        self.assertNotLogged("Re-emitting")
 
         framework.reemit()
 
@@ -368,6 +390,7 @@ class TestFramework(BaseTestCase):
         #    we'd get a foo=3).
         #
         self.assertEqual(obs.seen, ["on_foo:foo=2", "on_foo:foo=2"])
+        self.assertLoggedDebug("Re-emitting <MyEvent via MyNotifier[1]/foo[1]>.")
 
     def test_weak_observer(self):
         framework = self.create_framework()
@@ -771,6 +794,90 @@ class TestFramework(BaseTestCase):
             "{'bar': <class 'test.test_framework.TestFramework'>}")
         self.assertEqual(str(cm.exception), expected)
 
+    def test_unobserved_events_dont_leave_cruft(self):
+        class FooEvent(EventBase):
+            def snapshot(self):
+                return {'content': 1}
+
+        class Events(ObjectEvents):
+            foo = EventSource(FooEvent)
+
+        class Emitter(Object):
+            on = Events()
+
+        framework = self.create_framework()
+        e = Emitter(framework, 'key')
+        e.on.foo.emit()
+        ev_1_handle = Handle(e.on, "foo", "1")
+        with self.assertRaises(NoSnapshotError):
+            framework.load_snapshot(ev_1_handle)
+        # Committing will save the framework's state, but no other snapshots should be saved
+        framework.commit()
+        events = framework._storage.list_snapshots()
+        self.assertEqual(list(events), [framework._stored.handle.path])
+
+    def test_event_regex(self):
+        examples = [
+            'Ubuntu/on/config_changed[7]',
+            'on/commit[9]',
+            'on/pre_commit[8]',
+        ]
+        non_examples = [
+            'StoredStateData[_stored]',
+            'ObjectWithSTorage[obj]StoredStateData[_stored]',
+        ]
+        regex = re.compile(_event_regex)
+        for e in examples:
+            self.assertIsNotNone(regex.match(e))
+        for e in non_examples:
+            self.assertIsNone(regex.match(e))
+
+    def test_remove_unreferenced_events(self):
+        framework = self.create_framework()
+
+        class Evt(EventBase):
+            pass
+
+        class Events(ObjectEvents):
+            event = EventSource(Evt)
+
+        class ObjectWithStorage(Object):
+            _stored = StoredState()
+            on = Events()
+
+            def __init__(self, framework, key):
+                super().__init__(framework, key)
+                self._stored.set_default(foo=2)
+                self.framework.observe(self.on.event, self._on_event)
+
+            def _on_event(self, event):
+                event.defer()
+
+        # This is an event that 'happened in the past' that doesn't have an associated notice.
+        o = ObjectWithStorage(framework, 'obj')
+        handle = Handle(o.on, 'event', '100')
+        event = Evt(handle)
+        framework.save_snapshot(event)
+        self.assertEqual(list(framework._storage.list_snapshots()), [handle.path])
+        o.on.event.emit()
+        self.assertEqual(
+            list(framework._storage.notices('')),
+            [('ObjectWithStorage[obj]/on/event[1]', 'ObjectWithStorage[obj]', '_on_event')])
+        framework.commit()
+        self.assertEqual(
+            sorted(framework._storage.list_snapshots()),
+            sorted(['ObjectWithStorage[obj]/on/event[100]',
+                    'StoredStateData[_stored]',
+                    'ObjectWithStorage[obj]/StoredStateData[_stored]',
+                    'ObjectWithStorage[obj]/on/event[1]']))
+        framework.remove_unreferenced_events()
+        self.assertEqual(
+            sorted(framework._storage.list_snapshots()),
+            sorted([
+                'StoredStateData[_stored]',
+                'ObjectWithStorage[obj]/StoredStateData[_stored]',
+                'ObjectWithStorage[obj]/on/event[1]']))
+
 
 class TestStoredState(BaseTestCase):
 
@@ -1132,8 +1239,8 @@ class TestStoredState(BaseTestCase):
             _stored = StoredState()
 
         class WrappedFramework(Framework):
-            def __init__(self, data_path, charm_dir, meta, model):
-                super().__init__(data_path, charm_dir, meta, model)
+            def __init__(self, store, charm_dir, meta, model):
+                super().__init__(store, charm_dir, meta, model)
                 self.snapshots = []
 
             def save_snapshot(self, value):
@@ -1143,7 +1250,8 @@ class TestStoredState(BaseTestCase):
 
         # Validate correctness of modification operations.
         for get_a, b, expected_res, op, validate_op in test_operations:
-            framework = WrappedFramework(self.tmpdir / "framework.data", self.tmpdir, None, None)
+            storage = SQLiteStorage(self.tmpdir / "framework.data")
+            framework = WrappedFramework(storage, self.tmpdir, None, None)
             obj = SomeObject(framework, '1')
 
             obj._stored.a = get_a()
@@ -1168,8 +1276,8 @@ class TestStoredState(BaseTestCase):
             framework.commit()
             framework.close()
 
-            framework_copy = WrappedFramework(
-                self.tmpdir / "framework.data", self.tmpdir, None, None)
+            storage_copy = SQLiteStorage(self.tmpdir / "framework.data")
+            framework_copy = WrappedFramework(storage_copy, self.tmpdir, None, None)
 
             obj_copy2 = SomeObject(framework_copy, '1')
 
@@ -1372,6 +1480,10 @@ class GenericObserver(Object):
 @patch('sys.stderr', new_callable=io.StringIO)
 class BreakpointTests(BaseTestCase):
 
+    def setUp(self):
+        super().setUp()
+        logassert.setup(self, 'ops')
+
     def test_ignored(self, fake_stderr):
         # It doesn't do anything really unless proper environment is there.
         with patch.dict(os.environ):
@@ -1382,6 +1494,7 @@ class BreakpointTests(BaseTestCase):
             framework.breakpoint()
         self.assertEqual(mock.call_count, 0)
         self.assertEqual(fake_stderr.getvalue(), "")
+        self.assertNotLoggedWarning("Breakpoint", "skipped")
 
     def test_pdb_properly_called(self, fake_stderr):
         # The debugger needs to leave the user in the frame where the breakpoint is executed,
@@ -1509,9 +1622,19 @@ class BreakpointTests(BaseTestCase):
         # Some breakpoint was indicated, and the framework call used exactly that name.
         self.check_trace_set('mybreak', 'mybreak', 1)
 
-    def test_named_indicated_somethingelse(self, fake_stderr):
-        # Some breakpoint was indicated, but the framework call was not with that name.
+    def test_named_indicated_unnamed(self, fake_stderr):
+        # Some breakpoint was indicated, but the framework call was unnamed
         self.check_trace_set('some-breakpoint', None, 0)
+        self.assertLoggedWarning(
+            "Breakpoint None skipped",
+            "not found in the requested breakpoints: ['some-breakpoint']")
+
+    def test_named_indicated_somethingelse(self, fake_stderr):
+        # Some breakpoint was indicated, but the framework call was with a different name
+        self.check_trace_set('some-breakpoint', 'other-name', 0)
+        self.assertLoggedWarning(
+            "Breakpoint 'other-name' skipped",
+            "not found in the requested breakpoints: ['some-breakpoint']")
 
     def test_named_indicated_ingroup(self, fake_stderr):
         # A multiple breakpoint was indicated, and the framework call used a name among those.
