@@ -32,12 +32,6 @@ from test.test_helpers import fake_script, fake_script_calls
 class TestModel(unittest.TestCase):
 
     def setUp(self):
-        def restore_env(env):
-            os.environ.clear()
-            os.environ.update(env)
-        self.addCleanup(restore_env, os.environ.copy())
-
-        os.environ['JUJU_UNIT_NAME'] = 'myapp/0'
         self.harness = ops.testing.Harness(ops.charm.CharmBase, meta='''
             name: myapp
             provides:
@@ -49,7 +43,11 @@ class TestModel(unittest.TestCase):
             peers:
               db2:
                 interface: db2
+            resources:
+              foo: {type: file, filename: foo.txt}
+              bar: {type: file, filename: bar.txt}
         ''')
+        self.addCleanup(self.harness.cleanup)
         self.relation_id_db0 = self.harness.add_relation('db0', 'db')
         self.model = self.harness.model
 
@@ -57,9 +55,17 @@ class TestModel(unittest.TestCase):
         self.assertIs(self.model.app, self.model.unit.app)
         self.assertIsNone(self.model.name)
 
-    def test_model_name(self):
-        m = ops.model.Model('unit/0', ops.charm.CharmMeta(), self.harness._backend,
-                            model_name='default')
+    def test_unit_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.unit = object()
+
+    def test_app_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.app = object()
+
+    def test_model_name_from_backend(self):
+        self.harness.set_model_name('default')
+        m = ops.model.Model(ops.charm.CharmMeta(), self.harness._backend)
         self.assertEqual(m.name, 'default')
         with self.assertRaises(AttributeError):
             m.name = "changes-disallowed"
@@ -85,6 +91,10 @@ class TestModel(unittest.TestCase):
             ('relation_list', rel_app2),
         ])
 
+    def test_relations_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.relations = {}
+
     def test_get_relation(self):
         # one relation on db1
         # two relations on db0
@@ -211,6 +221,8 @@ class TestModel(unittest.TestCase):
                                    self.model.get_relation('db1').units))
         # Force memory cache to be loaded.
         self.assertIn('host', rel_db1.data[remoteapp1_0])
+        self.assertEqual(repr(rel_db1.data[remoteapp1_0]), "{'host': 'remoteapp1/0'}")
+
         with self.assertRaises(ops.model.RelationDataError):
             rel_db1.data[remoteapp1_0]['foo'] = 'bar'
         self.assertNotIn('foo', rel_db1.data[remoteapp1_0])
@@ -221,6 +233,16 @@ class TestModel(unittest.TestCase):
             ('relation_get', relation_id, 'remoteapp1/0', False),
         ])
 
+        # this will fire more backend calls
+        # the CountEqual and weird (and brittle) splitting is to accommodate python 3.5
+        # TODO: switch to assertEqual when we drop 3.5
+        self.assertCountEqual(
+            repr(rel_db1.data)[1:-1].split(', '),
+            ["<ops.model.Unit myapp/0>: {}",
+             "<ops.model.Application myapp>: {}",
+             "<ops.model.Unit remoteapp1/0>: {'host': 'remoteapp1/0'}",
+             "<ops.model.Application remoteapp1>: {'secret': 'cafedeadbeef'}"])
+
     def test_relation_data_modify_our(self):
         relation_id = self.harness.add_relation('db1', 'remoteapp1')
         self.harness.update_relation_data(relation_id, 'myapp/0', {'host': 'nothing'})
@@ -303,6 +325,26 @@ class TestModel(unittest.TestCase):
             ('relation_set', relation_id, 'host', '', False),
         ])
 
+    def test_relation_data_del_missing_key(self):
+        relation_id = self.harness.add_relation('db1', 'remoteapp1')
+        self.harness.update_relation_data(relation_id, 'myapp/0', {'host': 'bar'})
+        self.harness.add_relation_unit(relation_id, 'remoteapp1/0')
+        self.resetBackendCalls()
+
+        rel_db1 = self.model.get_relation('db1')
+        # Force memory cache to be loaded.
+        self.assertIn('host', rel_db1.data[self.model.unit])
+        rel_db1.data[self.model.unit]['port'] = ''   # Same as a delete, should not fail.
+        self.assertNotIn('port', rel_db1.data[self.model.unit])
+        self.assertEqual({'host': 'bar'}, self.harness.get_relation_data(relation_id, 'myapp/0'))
+
+        self.assertBackendCalls([
+            ('relation_ids', 'db1'),
+            ('relation_list', relation_id),
+            ('relation_get', relation_id, 'myapp/0', False),
+            ('relation_set', relation_id, 'port', '', False),
+        ])
+
     def test_relation_set_fail(self):
         relation_id = self.harness.add_relation('db1', 'remoteapp1')
         self.harness.update_relation_data(relation_id, 'myapp/0', {'host': 'myapp-0'})
@@ -373,6 +415,10 @@ class TestModel(unittest.TestCase):
 
         self.assertBackendCalls([('config_get',)])
 
+    def test_config_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.config = {}
+
     def test_is_leader(self):
         relation_id = self.harness.add_relation('db1', 'remoteapp1')
         self.harness.add_relation_unit(relation_id, 'remoteapp1/0')
@@ -415,69 +461,38 @@ class TestModel(unittest.TestCase):
         self.assertBackendCalls([])
 
     def test_resources(self):
-        # TODO: (jam) 2020-05-07 Harness doesn't yet support resource-get issue #262
-        meta = ops.charm.CharmMeta()
-        meta.resources = {'foo': None, 'bar': None}
-        model = ops.model.Model('myapp/0', meta, ops.model._ModelBackend())
+        with self.assertRaises(ops.model.ModelError):
+            self.harness.model.resources.fetch('foo')
+
+        self.harness.add_resource('foo', 'foo contents\n')
+        self.harness.add_resource('bar', '')
 
         with self.assertRaises(RuntimeError):
-            model.resources.fetch('qux')
+            self.harness.model.resources.fetch('qux')
 
-        fake_script(self, 'resource-get', 'exit 1')
-        with self.assertRaises(ops.model.ModelError):
-            model.resources.fetch('foo')
+        self.assertEqual(self.harness.model.resources.fetch('foo').name, 'foo.txt')
+        self.assertEqual(self.harness.model.resources.fetch('bar').name, 'bar.txt')
 
-        fake_script(self, 'resource-get',
-                    'echo /var/lib/juju/agents/unit-test-0/resources/$1/$1.tgz')
-        self.assertEqual(model.resources.fetch('foo').name, 'foo.tgz')
-        self.assertEqual(model.resources.fetch('bar').name, 'bar.tgz')
+    def test_resources_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.resources = object()
 
     def test_pod_spec(self):
-        # TODO: (jam) 2020-05-07 Harness doesn't yet expose pod-spec-set issue #261
-        meta = ops.charm.CharmMeta.from_yaml('''
-            name: myapp
-        ''')
-        model = ops.model.Model('myapp/0', meta, ops.model._ModelBackend())
-        fake_script(self, 'pod-spec-set', """
-                    cat $2 > $(dirname $0)/spec.json
-                    [[ -n $4 ]] && cat $4 > $(dirname $0)/k8s_res.json || true
-                    """)
-        fake_script(self, 'is-leader', 'echo true')
-        spec_path = self.fake_script_path / 'spec.json'
-        k8s_res_path = self.fake_script_path / 'k8s_res.json'
+        self.harness.set_leader(True)
+        self.harness.model.pod.set_spec({'foo': 'bar'})
+        self.assertEqual(self.harness.get_pod_spec(), ({'foo': 'bar'}, None))
 
-        def check_calls(calls):
-            # There may 1 or 2 calls because of is-leader.
-            self.assertLessEqual(len(fake_calls), 2)
-            pod_spec_call = next(filter(lambda c: c[0] == 'pod-spec-set', calls))
-            self.assertEqual(pod_spec_call[:2], ['pod-spec-set', '--file'])
+        self.harness.model.pod.set_spec({'bar': 'foo'}, {'qux': 'baz'})
+        self.assertEqual(self.harness.get_pod_spec(), ({'bar': 'foo'}, {'qux': 'baz'}))
 
-            # 8 bytes are used as of python 3.4.0, see Python bug #12015.
-            # Other characters are from POSIX 3.282 (Portable Filename
-            # Character Set) a subset of which Python's mkdtemp uses.
-            self.assertRegex(pod_spec_call[2], '.*/tmp[A-Za-z0-9._-]{8}-pod-spec-set')
-
-        model.pod.set_spec({'foo': 'bar'})
-        self.assertEqual(spec_path.read_text(), '{"foo": "bar"}')
-        self.assertFalse(k8s_res_path.exists())
-
-        fake_calls = fake_script_calls(self, clear=True)
-        check_calls(fake_calls)
-
-        model.pod.set_spec({'bar': 'foo'}, {'qux': 'baz'})
-        self.assertEqual(spec_path.read_text(), '{"bar": "foo"}')
-        self.assertEqual(k8s_res_path.read_text(), '{"qux": "baz"}')
-
-        fake_calls = fake_script_calls(self, clear=True)
-        check_calls(fake_calls)
-
-        # Create a new model to drop is-leader caching result.
-        self.backend = ops.model._ModelBackend()
-        meta = ops.charm.CharmMeta()
-        model = ops.model.Model('myapp/0', meta, self.backend)
-        fake_script(self, 'is-leader', 'echo false')
+        # no leader -> no set pod spec
+        self.harness.set_leader(False)
         with self.assertRaises(ops.model.ModelError):
-            model.pod.set_spec({'foo': 'bar'})
+            self.harness.model.pod.set_spec({'foo': 'bar'})
+
+    def test_pod_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.pod = object()
 
     def test_base_status_instance_raises(self):
         with self.assertRaises(TypeError):
@@ -651,7 +666,7 @@ class TestModel(unittest.TestCase):
         # TODO: (jam) 2020-05-07 Harness doesn't yet expose storage-get issue #263
         meta = ops.charm.CharmMeta()
         meta.storages = {'disks': None, 'data': None}
-        model = ops.model.Model('myapp/0', meta, ops.model._ModelBackend())
+        model = ops.model.Model(meta, ops.model._ModelBackend('myapp/0'))
 
         fake_script(self, 'storage-list', '''
             if [ "$1" = disks ]; then
@@ -706,6 +721,10 @@ class TestModel(unittest.TestCase):
             with self.assertRaises(TypeError):
                 model.storages.request('data', count_v)
 
+    def test_storages_immutable(self):
+        with self.assertRaises(AttributeError):
+            self.model.storages = {}
+
     def resetBackendCalls(self):
         self.harness._get_backend_calls(reset=True)
 
@@ -716,13 +735,6 @@ class TestModel(unittest.TestCase):
 class TestModelBindings(unittest.TestCase):
 
     def setUp(self):
-        def restore_env(env):
-            os.environ.clear()
-            os.environ.update(env)
-        self.addCleanup(restore_env, os.environ.copy())
-
-        os.environ['JUJU_UNIT_NAME'] = 'myapp/0'
-
         meta = ops.charm.CharmMeta()
         meta.relations = {
             'db0': RelationMeta(
@@ -732,8 +744,8 @@ class TestModelBindings(unittest.TestCase):
             'db2': RelationMeta(
                 RelationRole.peer, 'db2', {'interface': 'db2', 'scope': 'global'}),
         }
-        self.backend = ops.model._ModelBackend()
-        self.model = ops.model.Model('myapp/0', meta, self.backend)
+        self.backend = ops.model._ModelBackend('myapp/0')
+        self.model = ops.model.Model(meta, self.backend)
 
         fake_script(self, 'relation-ids',
                     """([ "$1" = db0 ] && echo '["db0:4"]') || echo '[]'""")
@@ -866,15 +878,12 @@ class TestModelBindings(unittest.TestCase):
 class TestModelBackend(unittest.TestCase):
 
     def setUp(self):
-        os.environ['JUJU_UNIT_NAME'] = 'myapp/0'
-        self.addCleanup(os.environ.pop, 'JUJU_UNIT_NAME')
-
         self._backend = None
 
     @property
     def backend(self):
         if self._backend is None:
-            self._backend = ops.model._ModelBackend()
+            self._backend = ops.model._ModelBackend('myapp/0')
         return self._backend
 
     def test_relation_get_set_is_app_arg(self):
@@ -897,7 +906,7 @@ class TestModelBackend(unittest.TestCase):
         meta = ops.charm.CharmMeta.from_yaml('''
             name: myapp
         ''')
-        model = ops.model.Model('myapp/0', meta, self.backend)
+        model = ops.model.Model(meta, self.backend)
         fake_script(self, 'is-leader', 'echo false')
         self.assertFalse(model.unit.is_leader())
 
@@ -915,6 +924,8 @@ class TestModelBackend(unittest.TestCase):
         self.assertTrue(model.unit.is_leader())
 
     def test_relation_tool_errors(self):
+        self.addCleanup(os.environ.pop, 'JUJU_VERSION', None)
+        os.environ['JUJU_VERSION'] = '2.8.0'
         err_msg = 'ERROR invalid value "$2" for option -r: relation not found'
 
         test_cases = [(
@@ -931,29 +942,79 @@ class TestModelBackend(unittest.TestCase):
             lambda: fake_script(self, 'relation-set', 'echo fooerror >&2 ; exit 1'),
             lambda: self.backend.relation_set(3, 'foo', 'bar', is_app=False),
             ops.model.ModelError,
-            [['relation-set', '-r', '3', 'foo=bar', '--app=False']],
+            [['relation-set', '-r', '3', 'foo=bar']],
         ), (
             lambda: fake_script(self, 'relation-set', 'echo {} >&2 ; exit 2'.format(err_msg)),
             lambda: self.backend.relation_set(3, 'foo', 'bar', is_app=False),
             ops.model.RelationNotFoundError,
-            [['relation-set', '-r', '3', 'foo=bar', '--app=False']],
+            [['relation-set', '-r', '3', 'foo=bar']],
+        ), (
+            lambda: None,
+            lambda: self.backend.relation_set(3, 'foo', 'bar', is_app=True),
+            ops.model.RelationNotFoundError,
+            [['relation-set', '-r', '3', 'foo=bar', '--app']],
         ), (
             lambda: fake_script(self, 'relation-get', 'echo fooerror >&2 ; exit 1'),
             lambda: self.backend.relation_get(3, 'remote/0', is_app=False),
             ops.model.ModelError,
-            [['relation-get', '-r', '3', '-', 'remote/0', '--app=False', '--format=json']],
+            [['relation-get', '-r', '3', '-', 'remote/0', '--format=json']],
         ), (
             lambda: fake_script(self, 'relation-get', 'echo {} >&2 ; exit 2'.format(err_msg)),
             lambda: self.backend.relation_get(3, 'remote/0', is_app=False),
             ops.model.RelationNotFoundError,
-            [['relation-get', '-r', '3', '-', 'remote/0', '--app=False', '--format=json']],
+            [['relation-get', '-r', '3', '-', 'remote/0', '--format=json']],
+        ), (
+            lambda: None,
+            lambda: self.backend.relation_get(3, 'remote/0', is_app=True),
+            ops.model.RelationNotFoundError,
+            [['relation-get', '-r', '3', '-', 'remote/0', '--app', '--format=json']],
         )]
 
-        for do_fake, run, exception, calls in test_cases:
-            do_fake()
-            with self.assertRaises(exception):
-                run()
-            self.assertEqual(fake_script_calls(self, clear=True), calls)
+        for i, (do_fake, run, exception, calls) in enumerate(test_cases):
+            with self.subTest(i):
+                do_fake()
+                with self.assertRaises(exception):
+                    run()
+                self.assertEqual(fake_script_calls(self, clear=True), calls)
+
+    def test_relation_get_juju_version_quirks(self):
+        self.addCleanup(os.environ.pop, 'JUJU_VERSION', None)
+
+        fake_script(self, 'relation-get', '''echo '{"foo": "bar"}' ''')
+
+        # on 2.7.0+, things proceed as expected
+        for v in ['2.8.0', '2.7.0']:
+            with self.subTest(v):
+                os.environ['JUJU_VERSION'] = v
+                rel_data = self.backend.relation_get(1, 'foo/0', is_app=True)
+                self.assertEqual(rel_data, {"foo": "bar"})
+                calls = [' '.join(i) for i in fake_script_calls(self, clear=True)]
+                self.assertEqual(calls, ['relation-get -r 1 - foo/0 --app --format=json'])
+
+        # before 2.7.0, it just fails (no --app support)
+        os.environ['JUJU_VERSION'] = '2.6.9'
+        with self.assertRaisesRegex(RuntimeError, 'not supported on Juju version 2.6.9'):
+            self.backend.relation_get(1, 'foo/0', is_app=True)
+        self.assertEqual(fake_script_calls(self), [])
+
+    def test_relation_set_juju_version_quirks(self):
+        self.addCleanup(os.environ.pop, 'JUJU_VERSION', None)
+
+        fake_script(self, 'relation-set', 'exit 0')
+
+        # on 2.7.0+, things proceed as expected
+        for v in ['2.8.0', '2.7.0']:
+            with self.subTest(v):
+                os.environ['JUJU_VERSION'] = v
+                self.backend.relation_set(1, 'foo', 'bar', is_app=True)
+                calls = [' '.join(i) for i in fake_script_calls(self, clear=True)]
+                self.assertEqual(calls, ['relation-set -r 1 foo=bar --app'])
+
+        # before 2.7.0, it just fails always (no --app support)
+        os.environ['JUJU_VERSION'] = '2.6.9'
+        with self.assertRaisesRegex(RuntimeError, 'not supported on Juju version 2.6.9'):
+            self.backend.relation_set(1, 'foo', 'bar', is_app=True)
+        self.assertEqual(fake_script_calls(self), [])
 
     def test_status_get(self):
         # taken from actual Juju output
@@ -1008,7 +1069,7 @@ class TestModelBackend(unittest.TestCase):
         meta = ops.charm.CharmMeta.from_yaml('''
             name: myapp
         ''')
-        model = ops.model.Model('myapp/0', meta, self.backend)
+        model = ops.model.Model(meta, self.backend)
         fake_script(self, 'status-set', 'exit 1')
         fake_script(self, 'is-leader', 'echo true')
 
